# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tokenizers.normalizers import Lowercase, NFKC, Sequence


class InputStringNormalizer(object):
    def __init__(self, do_lower_case=False, do_nfkc=False):
        self.do_lower_case: bool = do_lower_case
        self.do_nfkc: bool = do_nfkc
        self._normalizer: Sequence = self._init_normalizer()

    def _init_normalizer(self) -> Sequence:
        normalizers = []
        if self.do_lower_case:
            normalizers.append(Lowercase())
        if self.do_nfkc:
            normalizers.append(NFKC())
        return Sequence(normalizers)

    @property
    def normalizer(self) -> Sequence:
        return self._normalizer

    def normalize_str(self, text: str) -> str:
        return self.normalizer.normalize_str(text)



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from logzero import logger
from progressbar import progressbar
from typing import List

from corpus_preprocessing.filter.sentence_filter import (
    SF, SentenceFilterName,
    SequenceSentenceFilter,
    EmailFilter, UrlFilter, SequenceLengthFilter
)
from corpus_preprocessing.filter.document_filter import (
    DF, DocumentFilterName,
    SequenceDocumentFilter,
    NGWordsFilter, ShortDocumentFilter, ScriptFilter
)
from corpus_preprocessing.normalizer.sentence_normalizer import (
    SN, SentenceNormalizerName,
    SequenceSentenceNormalizer,
    CitationNormalizer, WhitespaceNormalizer
)

from corpus_preprocessing.normalizer.document_normalizer import (
    DN, DocumentNormalizerName,
    SequenceDocumentNormalizer,
    ConcatShortSentenceNormalizer
)


def load_dataset(input_dataset_path: str) -> List[List[str]]:
    documents = []
    with open(input_dataset_path, 'r', encoding='utf-8') as f:
        document = []
        for sentence in f:
            sentence = sentence.strip()
            if sentence != '':
                document.append(sentence)
            else:
                documents.append(document)
                document = []

    return documents


def load_sentence_filters(sentence_filter_names: List[str], **kwargs) -> List[SF]:
    sentence_filters = []
    for sentence_filter_name in sentence_filter_names:
        if sentence_filter_name == SentenceFilterName.EMAIL:
            sentence_filters.append(EmailFilter())
        elif sentence_filter_name == SentenceFilterName.URL:
            sentence_filters.append(UrlFilter())
        elif sentence_filter_name == SentenceFilterName.SEQUENCE_LENGTH:
            sentence_filters.append(SequenceLengthFilter(min_seq_len=kwargs['min_seq_len'],
                                                         max_seq_len=kwargs['max_seq_len']))
        else:
            raise ValueError('Invalid sentence filter name `{}`: {}'.format(
                sentence_filter_name, ','.join(map(str, SentenceFilterName)))
            )

    return sentence_filters


def load_document_filters(document_filter_names: List[str], **kwargs) -> List[DF]:
    document_filters = []
    for document_filter_name in document_filter_names:
        if document_filter_name == DocumentFilterName.SHORT_DOCUMENT:
            document_filters.append(ShortDocumentFilter(min_sentence_num=kwargs['min_sentence_num']))
        elif document_filter_name == DocumentFilterName.SCRIPT:
            document_filters.append(ScriptFilter())
        elif document_filter_name == DocumentFilterName.NG_WORDS:
            document_filters.append(NGWordsFilter(ng_words_file_path=kwargs['ng_words_file_path']))
        else:
            raise ValueError('Invalid document filter name `{}`: {}'.format(
                document_filter_name, ','.join(map(str, DocumentFilterName)))
            )

    return document_filters


def load_sentence_normalizers(sentence_normalizer_names: List[str], **kwargs) -> List[SN]:
    sentence_normalizers = []
    for sentence_normalizer_name in sentence_normalizer_names:
        if sentence_normalizer_name == SentenceNormalizerName.CITATION:
            sentence_normalizers.append(CitationNormalizer())
        elif sentence_normalizer_name == SentenceNormalizerName.WHITESPACE:
            sentence_normalizers.append(WhitespaceNormalizer())
        else:
            raise ValueError('Invalid sentence normalizer name `{}`: {}'.format(
                sentence_normalizer_name, ','.join(map(str, SentenceNormalizerName)))
            )

    return sentence_normalizers


def load_document_normalizers(document_normalizer_names: List[str], **kwargs) -> List[DN]:
    document_normalizers = []
    for document_normalizer_name in document_normalizer_names:
        if document_normalizer_name == DocumentNormalizerName.CONCAT_SHORT_SENTENCE:
            document_normalizers.append(ConcatShortSentenceNormalizer(concat_char_num=kwargs['concat_char_num']))
        else:
            raise ValueError('Invalid document normalizer name `{}`: {}'.format(
                document_normalizer_name, ','.join(map(str, DocumentNormalizerName)))
            )

    return document_normalizers


def main():
    args = get_args()
    for k, v in vars(args).items():
        logger.info('{}: {}'.format(k, v))

    sequence_sentence_filter = SequenceSentenceFilter(
        load_sentence_filters(args.sentence_filter_names,
                              min_seq_len=args.min_seq_len,
                              max_seq_len=args.max_seq_len)
    )

    sequence_document_filter = SequenceDocumentFilter(
        load_document_filters(args.document_filter_names,
                              min_sentence_num=args.min_sentence_num,
                              ng_words_file_path=args.ng_words_file_path)
    )

    sequence_sentence_normalizer = SequenceSentenceNormalizer(
        load_sentence_normalizers(args.sentence_normalizer_names)
    )

    sequence_document_normalizer = SequenceDocumentNormalizer(
        load_document_normalizers(args.document_normalizer_names,
                                  concat_char_num=args.concat_char_num)
    )

    documents = load_dataset(args.input_dataset_path)
    preprocessed_documents = []
    for document in progressbar(documents):
        # normalize
        document = [sequence_sentence_normalizer.normalize(s) for s in document]
        document = sequence_document_normalizer.normalize(document)

        # filter
        document = [s for s in document if not sequence_sentence_filter.is_filtered(s)]
        if not sequence_document_filter.is_filtered(document):
            preprocessed_documents.append(document)

    logger.info('#Document w/o filtering:\t{}'.format(len(documents)))
    logger.info('#Document w/ filtering:\t{}'.format(len(preprocessed_documents)))

    with open(args.output_dataset_path, 'w', encoding='utf-8') as f:
        f.write('\n\n'.join(['\n'.join(document) for document in preprocessed_documents]))


def get_args():
    parser = argparse.ArgumentParser(description='Cleaning and corpus_preprocessing dataset.')
    parser.add_argument('-i', '--input_dataset_path', required=True,
                        help='Input dataset.')
    parser.add_argument('-o', '--output_dataset_path', required=True,
                        help='Output dataset.')

    # Sentence filters
    parser.add_argument('-sf', '--sentence_filter_names', nargs='*', default=list(),
                        choices=SentenceFilterName, type=SentenceFilterName,
                        help='A list of filter names to remove unnecessary sentences from the training corpus.')
    parser.add_argument('--min_seq_len', type=int, default=10,
                        help='The minimum number of characters a sentence should contain (for SequenceLengthFilter).')
    parser.add_argument('--max_seq_len', type=int, default=200,
                        help='The maximum number of characters a sentence should contain (for SequenceLengthFilter).')

    # Document filters
    parser.add_argument('-df', '--document_filter_names', nargs='*', default=list(),
                        choices=DocumentFilterName, type=DocumentFilterName,
                        help='A list of filter names to remove unnecessary documents from the training corpus.')
    parser.add_argument('--min_sentence_num', type=int, default=5,
                        help='The minimum number of sentences a document should contain (for ShortDocumentFilter).')
    parser.add_argument('--ng_words_file_path',
                        default='./resources/ng_words.txt',
                        help='A file path of NG word list (for NGWordsFilter).')

    # Sentence normalizers
    parser.add_argument('-sn', '--sentence_normalizer_names', nargs='*', default=list(),
                        choices=SentenceNormalizerName, type=SentenceNormalizerName,
                        help='A list of filter names to normalize sentences from the training corpus.')

    # Document normalizers
    parser.add_argument('-dn', '--document_normalizer_names', nargs='*', default=list(),
                        choices=DocumentNormalizerName, type=DocumentNormalizerName,
                        help='A list of filter names to normalize documents from the training corpus.')
    parser.add_argument('--concat_char_num', type=int, default=2,
                        help='The maximum number of characters to be concatenated with the previous sentence '
                             '(for ConcatShortSentenceNormalizer).')

    args = parser.parse_args()

    return args


if __name__ == '__main__':
    main()



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import List, TypeVar


class SentenceNormalizer(object):

    def normalize(self, sentence: str) -> str:
        """
        Normalizes the sentence.

        Args:
            sentence (str): A sentence.

        Returns:
            str: A normalized sentence.
        """
        raise NotImplementedError()


SN = TypeVar('SN', bound=SentenceNormalizer)


class SequenceSentenceNormalizer(SentenceNormalizer):
    def __init__(self, sentence_normalizers: List[SN]):
        """
        Constructs a SequenceSentenceNormalizer.

        Args:
            sentence_normalizers (List[SN]): A list of SentenceNormalizers.
        """
        self._sentence_normalizers: List[SN] = sentence_normalizers

    def normalize(self, sentence: str) -> str:
        """
        Applies normalizers to the sentence in sequence.

        Args:
            sentence (str): A sentence.

        Returns:
            str: A normalized sentence.
        """
        for sentence_normalizer in self._sentence_normalizers:
            sentence = sentence_normalizer.normalize(sentence)

        return sentence


class WhitespaceNormalizer(SentenceNormalizer):
    continuous_whitespace_pattern = re.compile(r'\s+')

    def normalize(self, sentence: str) -> str:
        """
        Removes invisible characters and replaces consecutive whitespace with a single whitespace.

        Args:
            sentence (str): A sentence.

        Returns:
            str: A sentence with consecutive whitespace and invisible characters removed.
        """
        sentence = "".join(c for c in sentence if c.isprintable())
        sentence = self.continuous_whitespace_pattern.sub(' ', sentence)

        return sentence


class CitationNormalizer(SentenceNormalizer):
    citation_pattern = re.compile(r'\[\d+?\]|\[要.+?\]|\{\{+[^{}]+?\}\}+|\[(要出典|リンク切れ|.+?\?)\]')

    def normalize(self, sentence: str) -> str:
        """
        Removes citation markers.

        Args:
            sentence (str): A sentence.

        Returns:
            str: A sentence with citation markers removed.
        """
        return self.citation_pattern.sub('', sentence)



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
from collections import OrderedDict
from typing import Dict, List, Optional, Tuple

from sudachipy.morpheme import Morpheme
from transformers.models.bert_japanese.tokenization_bert_japanese import CharacterTokenizer
from transformers.models.bert.tokenization_bert import WordpieceTokenizer
from transformers.tokenization_utils import PreTrainedTokenizer

from .input_string_normalizer import InputStringNormalizer
from .sudachipy_word_tokenizer import SudachipyWordTokenizer
from .word_formatter import word_formatter


VOCAB_FILES_NAMES = {"vocab_file": "vocab.txt"}

# TODO: set official URL
PRETRAINED_VOCAB_FILES_MAP = {
    "vocab_file": {
        "WorksApplications/bert-base-japanese-sudachi": "https://.../vocab.txt",
    }
}

PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES = {
    "WorksApplications/bert-base-japanese-sudachi": 512,
}

PRETRAINED_INIT_CONFIGURATION = {
    "WorksApplications/bert-base-japanese-sudachi": {
        "do_lower_case": False,
        "word_tokenizer_type": "sudachipy",
        "subword_tokenizer_type": "pos_substitution",
    },
}


def load_vocabulary(vocab_file: str = VOCAB_FILES_NAMES["vocab_file"]) -> Dict[str, int]:
    """
    Loads a vocabulary file into a dictionary.

    Args:
        vocab_file (str): Vocabulary file path.

    Returns:
        Dict[str, int]: Dictionary of vocabulary and its index.
    """
    vocab = OrderedDict()
    with open(vocab_file, "r", encoding="utf-8") as reader:
        tokens = reader.readlines()
    for index, token in enumerate(tokens):
        token = token.rstrip("\n").split("\t")[0]
        vocab[token] = index
    return vocab


def save_vocabulary(vocab: Dict[str, int], save_directory: str, filename_prefix: Optional[str] = None,
                    vocab_file: str = VOCAB_FILES_NAMES["vocab_file"]) -> Tuple[str]:
    """
    Save the vocabulary.

    Args:
        vocab (Dict[str, int]): Dictionary of vocabulary and its index.
        save_directory (str): The output directory where the vocabulary will be stored.
        filename_prefix (str | None): The filename prefix of the vocabulary file.
        vocab_file (str): The filename of the vocabulary file.

    Returns:
        Tuple[str]: Saved vocabulary path.
    """
    index = 0
    if os.path.isdir(save_directory):
        vocab_file = os.path.join(
            save_directory, (filename_prefix + "-" if filename_prefix else "") + vocab_file
        )
    else:
        vocab_file = (filename_prefix + "-" if filename_prefix else "") + save_directory
    with open(vocab_file, "w", encoding="utf-8") as writer:
        for token, token_index in sorted(vocab.items(), key=lambda kv: kv[1]):
            assert index == token_index, f"Error {vocab_file}: vocabulary indices are not consecutive, '{token}' {index} != {token_index}."
            writer.write(token + "\n")
            index += 1
    return (vocab_file,)


SUBWORD_TOKENIZER_TYPES = [
    "pos_substitution",
    "wordpiece",
    "character",
]


def pos_substitution_format(token: Morpheme) -> str:
    """
    Creates POS tag by combining each POS.

    Args:
        token (Morpheme): Morpheme in a sentence.

    Returns:
        str: POS tag.
    """
    hierarchy = token.part_of_speech()
    pos = [hierarchy[0]]
    for p in hierarchy[1:]:
        if p == "*":
            break
        pos.append(p)

    return "[{}]".format("-".join(pos))


class BertSudachipyTokenizer(PreTrainedTokenizer):

    vocab_files_names = VOCAB_FILES_NAMES
    pretrained_vocab_files_map = PRETRAINED_VOCAB_FILES_MAP
    pretrained_init_configuration = PRETRAINED_INIT_CONFIGURATION
    max_model_input_sizes = PRETRAINED_POSITIONAL_EMBEDDINGS_SIZES

    def __init__(
            self,
            vocab_file,
            do_lower_case=False,
            do_nfkc=False,
            do_word_tokenize=True,
            do_subword_tokenize=True,
            word_tokenizer_type="sudachipy",
            subword_tokenizer_type="wordpiece",
            unk_token="[UNK]",
            sep_token="[SEP]",
            pad_token="[PAD]",
            cls_token="[CLS]",
            mask_token="[MASK]",
            word_form_type="surface",
            sudachipy_kwargs=None,
            **kwargs
    ):
        super().__init__(
            do_lower_case=do_lower_case,
            do_nfkc=do_nfkc,
            do_word_tokenize=do_word_tokenize,
            do_subword_tokenize=do_subword_tokenize,
            word_tokenizer_type=word_tokenizer_type,
            subword_tokenizer_type=subword_tokenizer_type,
            unk_token=unk_token,
            sep_token=sep_token,
            pad_token=pad_token,
            cls_token=cls_token,
            mask_token=mask_token,
            word_form_type=word_form_type,
            sudachipy_kwargs=sudachipy_kwargs,
            **kwargs,
        )

        if not os.path.isfile(vocab_file):
            raise ValueError(f"Can't find a vocabulary file at path '{vocab_file}'.")

        self.vocab = load_vocabulary(vocab_file)
        self.ids_to_tokens = OrderedDict([(ids, tok) for tok, ids in self.vocab.items()])

        if not do_word_tokenize:
            raise ValueError(f"`do_word_tokenize` must be True.")
        elif word_tokenizer_type != "sudachipy":
            raise ValueError(f"Invalid word_tokenizer_type '{word_tokenizer_type}' is specified.")

        self.lower_case = do_lower_case
        self.nfkc = do_nfkc
        self.normalizer = InputStringNormalizer(do_lower_case=self.do_lower_case, do_nfkc=self.nfkc)

        self.sudachipy_kwargs = copy.deepcopy(sudachipy_kwargs)
        self.word_tokenizer = SudachipyWordTokenizer(**(self.sudachipy_kwargs or {}))
        self.word_form_type = word_form_type
        self.word_formatter = word_formatter(self.word_form_type, self.word_tokenizer.sudachi_dict)

        self.do_subword_tokenize = do_subword_tokenize
        self.subword_tokenizer_type = subword_tokenizer_type
        if do_subword_tokenize:
            if subword_tokenizer_type == "pos_substitution":
                self.subword_tokenizer = None
            elif subword_tokenizer_type == "wordpiece":
                self.subword_tokenizer = WordpieceTokenizer(vocab=self.vocab, unk_token=self.unk_token)
            elif subword_tokenizer_type == "character":
                self.subword_tokenizer = CharacterTokenizer(vocab=self.vocab, unk_token=self.unk_token)
            else:
                raise ValueError(f"Invalid subword_tokenizer_type '{subword_tokenizer_type}' is specified.")

    @property
    def do_lower_case(self):
        return self.lower_case

    @property
    def do_nfkc(self):
        return self.nfkc

    @property
    def vocab_size(self):
        return len(self.vocab)

    def get_vocab(self):
        return dict(self.vocab, **self.added_tokens_encoder)

    # TODO: need to investigate the serialization behavior
    def __getstate__(self):
        state = dict(self.__dict__)
        del state["word_tokenizer"]
        del state["word_formatter"]
        return state

    # TODO: need to investigate the serialization behavior
    def __setstate__(self, state):
        self.__dict__ = state
        self.word_tokenizer = SudachipyWordTokenizer(**(self.sudachipy_kwargs or {}))
        self.word_formatter = word_formatter(self.word_form_type, self.word_tokenizer.sudachi_dict)

    def _tokenize(self, text, **kwargs):
        text = self.normalizer.normalize_str(text)
        tokens = self.word_tokenizer.tokenize(text)

        if self.do_subword_tokenize:
            if self.subword_tokenizer_type == "pos_substitution":
                def _substitution(token):
                    word = self.word_formatter(token)
                    if word in self.vocab:
                        return word
                    substitute = pos_substitution_format(token)
                    if substitute in self.vocab:
                        return substitute
                    return self.unk_token
                split_tokens = [_substitution(token) for token in tokens]
            else:
                split_tokens = [sub_token for token in tokens for sub_token in self.subword_tokenizer.tokenize(
                    self.word_formatter(token)
                )]
        else:
            split_tokens = [self.word_formatter(token) for token in tokens]

        return split_tokens

    # <-- ported from BertTokenizer

    def _convert_token_to_id(self, token):
        """Converts a token (str) in an id using the vocab."""
        return self.vocab.get(token, self.vocab.get(self.unk_token))

    def _convert_id_to_token(self, index):
        """Converts an index (integer) in a token (str) using the vocab."""
        return self.ids_to_tokens.get(index, self.unk_token)

    def convert_tokens_to_string(self, tokens):
        """Converts a sequence of tokens (string) in a single string."""
        out_string = " ".join(tokens).replace(" ##", "").strip()
        return out_string

    def build_inputs_with_special_tokens(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Build model inputs from a sequence or a pair of sequence for sequence classification tasks by concatenating and
        adding special tokens. A BERT sequence has the following format:

        - single sequence: ``[CLS] X [SEP]``
        - pair of sequences: ``[CLS] A [SEP] B [SEP]``

        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs to which the special tokens will be added.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.

        Returns:
            :obj:`List[int]`: List of `input IDs <../glossary.html#input-ids>`__ with the appropriate special tokens.
        """
        if token_ids_1 is None:
            return [self.cls_token_id] + token_ids_0 + [self.sep_token_id]
        cls = [self.cls_token_id]
        sep = [self.sep_token_id]
        return cls + token_ids_0 + sep + token_ids_1 + sep

    def get_special_tokens_mask(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None, already_has_special_tokens: bool = False
    ) -> List[int]:
        """
        Retrieve sequence ids from a token list that has no special tokens added. This method is called when adding
        special tokens using the tokenizer ``prepare_for_model`` method.

        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.
            already_has_special_tokens (:obj:`bool`, `optional`, defaults to :obj:`False`):
                Whether or not the token list is already formatted with special tokens for the model.

        Returns:
            :obj:`List[int]`: A list of integers in the range [0, 1]: 1 for a special token, 0 for a sequence token.
        """

        if already_has_special_tokens:
            return super().get_special_tokens_mask(
                token_ids_0=token_ids_0, token_ids_1=token_ids_1, already_has_special_tokens=True
            )

        if token_ids_1 is not None:
            return [1] + ([0] * len(token_ids_0)) + [1] + ([0] * len(token_ids_1)) + [1]
        return [1] + ([0] * len(token_ids_0)) + [1]

    def create_token_type_ids_from_sequences(
        self, token_ids_0: List[int], token_ids_1: Optional[List[int]] = None
    ) -> List[int]:
        """
        Create a mask from the two sequences passed to be used in a sequence-pair classification task. A BERT sequence
        pair mask has the following format:

        ::

            0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1
            | first sequence    | second sequence |

        If :obj:`token_ids_1` is :obj:`None`, this method only returns the first portion of the mask (0s).

        Args:
            token_ids_0 (:obj:`List[int]`):
                List of IDs.
            token_ids_1 (:obj:`List[int]`, `optional`):
                Optional second list of IDs for sequence pairs.

        Returns:
            :obj:`List[int]`: List of `token type IDs <../glossary.html#token-type-ids>`_ according to the given
            sequence(s).
        """
        sep = [self.sep_token_id]
        cls = [self.cls_token_id]
        if token_ids_1 is None:
            return len(cls + token_ids_0 + sep) * [0]
        return len(cls + token_ids_0 + sep) * [0] + len(token_ids_1 + sep) * [1]

    # ported from BertTokenizer -->

    def save_vocabulary(self, save_directory: str, filename_prefix: Optional[str] = None,
                        vocab_file_name: str = VOCAB_FILES_NAMES["vocab_file"]) -> Tuple[str]:
        """
        Save the vocabulary.

        Args:
            save_directory (str): The output directory where the vocabulary will be stored.
            filename_prefix (str | None): The filename prefix of the vocabulary file.
            vocab_file_name (str): The filename of the vocabulary file.

        Returns:
            Tuple[str]: Saved vocabulary path.
        """
        return save_vocabulary(self.vocab, save_directory, filename_prefix, vocab_file_name)



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import List, TypeVar


class SentenceFilter(object):

    def is_filtered(self, sentence: str) -> bool:
        """
        Determine if the input sentence should be filtered or not.

        Args:
            sentence (str): A sentence of a document in a training corpus to pretrain model.

        Returns:
            bool: `True` if the sentence should be filtered, otherwise `False`.
        """
        raise NotImplementedError()


SF = TypeVar('SF', bound=SentenceFilter)


class SequenceSentenceFilter(SentenceFilter):
    def __init__(self, sentence_filters: List[SF]):
        """
        Constructs a SequenceSentenceFilter.

        Args:
            sentence_filters (List[SF]): A list of SentenceFilters.
        """
        self._sentence_filters: List[SF] = sentence_filters

    def is_filtered(self, sentence: str) -> bool:
        """
        Applies filters to the sentence in sequence.

        Args:
            sentence (str): A sentence.

        Returns:
            bool:`True` if the sentence should be filtered, otherwise `False`.
        """
        return any([sf.is_filtered(sentence) for sf in self._sentence_filters])


class UrlFilter(SentenceFilter):
    url_pattern = re.compile(r'(https?|sftp?)://[\w/:%#\$&\?\(\)~\.=\+\-]+')

    def is_filtered(self, sentence: str) -> bool:
        """
        Determines if the sentence contains URL.

        Args:
            sentence (str): A sentence.

        Returns:
            bool: `True` if the sentence contains URL, otherwise `False`.
        """
        return bool(self.url_pattern.search(sentence))


class EmailFilter(SentenceFilter):
    mail_pattern = re.compile(r'[\w\d_-]+@[\w\d_-]+\.[\w\d._-]+')

    def is_filtered(self, sentence: str) -> bool:
        """
        Determines if the sentence contains email address.

        Args:
            sentence (str): A sentence.

        Returns:
            bool: `True` if the sentence contains email address, otherwise `False`.
        """
        return bool(self.mail_pattern.search(sentence))


class SequenceLengthFilter(SentenceFilter):

    def __init__(self, min_seq_len: int = 10, max_seq_len: int = 200):
        """
        Constructs a SequenceLengthFilter.

        Args:
            min_seq_len (int): The minimum number of characters a sentence should contain.
            max_seq_len (int): The maximum number of characters a sentence should contain.
        """
        self.min_seq_len = min_seq_len
        self.max_seq_len = max_seq_len

    def is_filtered(self, sentence: str) -> bool:
        """
        Determines if the number of characters in the sentence is suitable.

        Args:
            sentence (str): A sentence.

        Returns:
            bool: `True` if the sentence length is either too short or too long, otherwise `False`.
        """
        return len(sentence) < self.min_seq_len or self.max_seq_len < len(sentence)



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from sudachipy.morpheme import Morpheme
from sudachipy.sudachipy import Dictionary

from .word_formatter import CONJUGATIVE_POS


class ConjugationPreservingNormalizer:
    def __init__(self, inflection_table_path, conjugation_type_table_path, sudachi_dict: Dictionary) -> None:
        """
        Constructs a ConjugationPreservingNormalizer.

        Args:
            inflection_table_path (:obj:`str`):
                In this table, get the difference between the inflected form to the end form each conjugation.
            conjugation_type_table_path (:obj:`str`):
                In this table, the conjugation types such as potential verb
                that change due to the normalization of the Sudachi dictionary are stored in blacklist format.
            sudachi_dict (:obj:`Dictionary`):
                dictionary instance.
        """
        self.sudachi_dict = sudachi_dict
        self.id2pos = list(enumerate(self.sudachi_dict.pos_matcher([()])))
        self.pos2id = {pos: pos_id for (pos_id, pos) in self.id2pos}
        self.is_necessary_inflection = sudachi_dict.pos_matcher(
            lambda p: p[0] in CONJUGATIVE_POS and (p[4] == "サ行変格" or p[5] != "終止形-一般")
        )  # 「為る->する」のため、サ行変格のみ 終止形-一般も変化

        with open(inflection_table_path) as jf:
            self.inflection_table = self._load_json(jf, "inflection")
        with open(conjugation_type_table_path) as jf:
            self.conjugation_type_table = self._load_json(jf, "conjugation_type")
    
    def _load_json(self, json_file, table_type) -> dict:
        """
        Convert json to python file when loading.

        Args:
            json_file (:obj:`TextIOWrapper`):
                json file object.
            table_type (:obj:`str`):
                Select "inflection" or "conjugation_type" and load in each format.

        Returns:
            Dict:
                inflection: Key of int and Inflection table of list.
                conjugation_type: Key of tuple and Conjugation type conversion destination of str.
        """
        if table_type not in ['inflection', 'conjugation_type']:
            raise ValueError('Invalid table_type error : {}'.format(table_type))

        data = json.load(json_file)
        table = {}
        for pos_0, convert_tables in data.items():
            for key, convert_table in convert_tables.items():
                if table_type == "inflection":
                    pos_4, pos_5 = key.split("|")
                    for pos in self.sudachi_dict.pos_matcher(lambda p: p[0] == pos_0 and p[4] == pos_4 and p[5] == pos_5):
                        pos_id = self.pos2id[pos]
                        table[pos_id] = convert_table
                elif table_type == "conjugation_type":
                    surface_token, reading, normalized_token, pos_4 = key.split("|")
                    for pos in self.sudachi_dict.pos_matcher(lambda p: p[0] == pos_0 and p[4] == pos_4):
                        pos_id = self.pos2id[pos]
                        table[(pos_id, surface_token, reading, normalized_token)] = convert_table

        return table

    def _change_pos(self, key: tuple, pos: tuple) -> tuple:
        """
        Make conjugation-type changes.

        If the part of speech does not exist after the change, it will be grouped into "一般". (Example: 撥音便, etc.)

        Args:
            key (:obj:`tuple`): (pos_id, surface, reading_form, normalized_form)
            pos (:obj:`tuple`): Get the part of speech.

        Returns:
            tuple: Changed part of speech
        """
        conj_type = self.conjugation_type_table[key]
        res = (*pos[:4], conj_type, pos[5])
        if res not in self.pos2id:
            conj_form = pos[5].split("-")[0] + "-一般"
            res = (*pos[:4], conj_type, conj_form)
            if res not in self.pos2id:
                res = (pos[0], "一般", pos[2], pos[3], conj_type, conj_form)

        assert res in self.pos2id
        return res
    
    def _is_changed_conjugation_type(self, key: tuple) -> bool:
        """
        Check if the conjugation type changes after being normalized.

        Args:
            key (:obj:`tuple`): (pos_id, surface, reading_form, normalized_form)

        Returns: bool
        """
        return key in self.conjugation_type_table

    def normalized(self, morpheme: Morpheme) -> str:
        """
        The output token retain conjugation information in word normalization by Sudachi tokenizer

        Args:
            morpheme (:obj:`Morpheme`): A Morpheme obtained from the analysis by sudachipy.

        Returns:
            str: Normalized token with conjugate information retained.
        """
        normalized_token = morpheme.normalized_form()
        if not self.is_necessary_inflection(morpheme):
            return normalized_token

        pos_id = morpheme.part_of_speech_id()
        conj_type_table_key = (pos_id, morpheme.surface(), morpheme.reading_form(), normalized_token)
        if self._is_changed_conjugation_type(conj_type_table_key):
            pos = self._change_pos(conj_type_table_key, morpheme.part_of_speech())
            if pos[5] == "終止形-一般":
                return normalized_token
            pos_id = self.pos2id[pos]

        for convert_table in self.inflection_table[pos_id]:
            if convert_table == ['', '']:
                return normalized_token
            if normalized_token.endswith(convert_table[0]):
                src = convert_table[0][::-1]
                tgt = convert_table[1][::-1]
                return normalized_token[::-1].replace(src, tgt, 1)[::-1]

        return normalized_token



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Dict, Iterator, List, Optional, TypeVar, Union

from logzero import logger
from tokenizers import Tokenizer, AddedToken, decoders, trainers
from tokenizers.models import WordPiece
from tokenizers.pre_tokenizers import BertPreTokenizer, PreTokenizer
from tokenizers.processors import BertProcessing
from tokenizers.implementations import BertWordPieceTokenizer
from tokenizers.implementations.base_tokenizer import BaseTokenizer

from .sudachipy_pretokenizer import CustomPreTokenizer
from ..input_string_normalizer import InputStringNormalizer


CPT = TypeVar('CPT', bound=CustomPreTokenizer)


class JapaneseBertWordPieceTokenizer(BaseTokenizer):
    def __init__(
            self,
            vocab: Optional[Union[str, Dict[str, int]]] = None,
            unk_token: Union[str, AddedToken] = "[UNK]",
            sep_token: Union[str, AddedToken] = "[SEP]",
            cls_token: Union[str, AddedToken] = "[CLS]",
            pad_token: Union[str, AddedToken] = "[PAD]",
            mask_token: Union[str, AddedToken] = "[MASK]",
            do_lower_case: bool = False,
            do_nfkc: bool = False,
            wordpieces_prefix: str = "##",
    ):
        if vocab is not None:
            tokenizer = Tokenizer(WordPiece(vocab, unk_token=str(unk_token)))
        else:
            tokenizer = Tokenizer(WordPiece(unk_token=str(unk_token)))

        # Let the tokenizer know about special tokens if they are part of the vocab
        if tokenizer.token_to_id(str(unk_token)) is not None:
            tokenizer.add_special_tokens([str(unk_token)])
        if tokenizer.token_to_id(str(sep_token)) is not None:
            tokenizer.add_special_tokens([str(sep_token)])
        if tokenizer.token_to_id(str(cls_token)) is not None:
            tokenizer.add_special_tokens([str(cls_token)])
        if tokenizer.token_to_id(str(pad_token)) is not None:
            tokenizer.add_special_tokens([str(pad_token)])
        if tokenizer.token_to_id(str(mask_token)) is not None:
            tokenizer.add_special_tokens([str(mask_token)])

        _normalizer = InputStringNormalizer(do_lower_case=do_lower_case, do_nfkc=do_nfkc)
        tokenizer.normalizer = _normalizer.normalizer
        tokenizer.pre_tokenizer = BertPreTokenizer()

        if vocab is not None:
            sep_token_id = tokenizer.token_to_id(str(sep_token))
            if sep_token_id is None:
                raise TypeError("sep_token not found in the vocabulary")
            cls_token_id = tokenizer.token_to_id(str(cls_token))
            if cls_token_id is None:
                raise TypeError("cls_token not found in the vocabulary")

            tokenizer.post_processor = BertProcessing(
                (str(sep_token), sep_token_id), (str(cls_token), cls_token_id)
            )
        tokenizer.decoder = decoders.WordPiece(prefix=wordpieces_prefix)

        parameters = {
            "model": "BertSudachiWordPiece",
            "unk_token": unk_token,
            "sep_token": sep_token,
            "cls_token": cls_token,
            "pad_token": pad_token,
            "mask_token": mask_token,
            "do_lower_case": do_lower_case,
            "do_nfkc": do_nfkc,
            "wordpieces_prefix": wordpieces_prefix,
        }

        super().__init__(tokenizer, parameters)

    @staticmethod
    def from_file(vocab: str, **kwargs):
        vocab = WordPiece.read_file(vocab)
        return BertWordPieceTokenizer(vocab, **kwargs)

    def train(
            self,
            files: Union[str, List[str]],
            vocab_size: int = 30000,
            min_frequency: int = 2,
            limit_alphabet: int = 1000,
            initial_alphabet: List[str] = [],
            special_tokens: List[Union[str, AddedToken]] = [
                "[PAD]",
                "[UNK]",
                "[CLS]",
                "[SEP]",
                "[MASK]",
            ],
            show_progress: bool = True,
            wordpieces_prefix: str = "##",
    ):
        """ Train the model using the given files """
        os.environ["TOKENIZERS_PARALLELISM"] = "false"

        logger.info("Parameters for training")
        logger.info("\tdo_lower_case: {}".format(self._parameters["do_lower_case"]))
        logger.info("\tdo_nfkc: {}".format(self._parameters["do_nfkc"]))
        logger.info("\tvocab_size: {}".format(vocab_size))
        logger.info("\tmin_frequency: {}".format(min_frequency))
        logger.info("\tlimit_alphabet: {}".format(limit_alphabet))
        logger.info("\tinitial_alphabet: {}".format(",".join(initial_alphabet)))
        logger.info("\tspecial_tokens: {}".format(",".join(special_tokens)))
        logger.info("\twordpieces_prefix: {}".format(wordpieces_prefix))

        trainer = trainers.WordPieceTrainer(
            vocab_size=vocab_size,
            min_frequency=min_frequency,
            limit_alphabet=limit_alphabet,
            initial_alphabet=initial_alphabet,
            special_tokens=special_tokens,
            show_progress=show_progress,
            continuing_subword_prefix=wordpieces_prefix,
        )
        if isinstance(files, str):
            files = [files]

        logger.info("Input files")
        logger.info("\n".join(map(lambda x: "\t{}".format(x), files)))

        self._tokenizer.train(files, trainer=trainer)
        logger.info("#Vocab: {}".format(self.get_vocab_size()))

    def train_from_iterator(
            self,
            iterator: Union[Iterator[str], Iterator[Iterator[str]]],
            vocab_size: int = 30000,
            min_frequency: int = 2,
            limit_alphabet: int = 1000,
            initial_alphabet: List[str] = [],
            special_tokens: List[Union[str, AddedToken]] = [
                "[PAD]",
                "[UNK]",
                "[CLS]",
                "[SEP]",
                "[MASK]",
            ],
            show_progress: bool = True,
            wordpieces_prefix: str = "##",
    ):
        """ Train the model using the given iterator """
        os.environ["TOKENIZERS_PARALLELISM"] = "false"

        logger.info("Parameters for training")
        logger.info("\tvocab_size: {}".format(vocab_size))
        logger.info("\tdo_lower_case: {}".format(self._parameters["do_lower_case"]))
        logger.info("\tdo_nfkc: {}".format(self._parameters["do_nfkc"]))
        logger.info("\tmin_frequency: {}".format(min_frequency))
        logger.info("\tlimit_alphabet: {}".format(limit_alphabet))
        logger.info("\tinitial_alphabet: {}".format(",".join(initial_alphabet)))
        logger.info("\tspecial_tokens: {}".format(",".join(special_tokens)))
        logger.info("\twordpieces_prefix: {}".format(wordpieces_prefix))

        trainer = trainers.WordPieceTrainer(
            vocab_size=vocab_size,
            min_frequency=min_frequency,
            limit_alphabet=limit_alphabet,
            initial_alphabet=initial_alphabet,
            special_tokens=special_tokens,
            show_progress=show_progress,
            continuing_subword_prefix=wordpieces_prefix,
        )
        self._tokenizer.train_from_iterator(iterator, trainer=trainer)

        logger.info("#Vocab: {}".format(self.get_vocab_size()))

    def set_pre_tokenizer(self, custom_pre_tokenizer: CPT):
        """
        Sets the custom tokenizer as pre-tokenizer.

        Args:
            custom_pre_tokenizer (CPT): Custom tokenizer that implements `custom_split`.
        """
        self.pre_tokenizer = PreTokenizer.custom(custom_pre_tokenizer)

    def save(self, output_tokenizer_path: str, pretty: bool = False):
        """
        Saves a config file of the tokenizer.

        Notes: To serialize tokenizer, puts dummy pre-tokenizer.

        Args:
            output_tokenizer_path (str): Output file path to be saved a config file of tokenizer.
            pretty (bool): Json format type.
        """
        logger.info("Saving config to `{}`".format(output_tokenizer_path))

        self.pre_tokenizer = BertPreTokenizer()  # dummy
        super().save(output_tokenizer_path, pretty=pretty)

    def save_vocab(self, output_dir: str, prefix: str):
        """
        Save the vocabulary.

        Args:
            output_dir (str): The path to the target directory in which to save the various files.
            prefix (str): An optional prefix, used to prefix each file name
        """
        logger.info("Saving vocab to `{}`".format(os.path.join(output_dir, "{}-vocab.txt".format(prefix))))

        self.model.save(output_dir, prefix)



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from re import Match
from sudachipy import Dictionary, SplitMode
from typing import List, TypeVar


class DocumentFilter(object):

    def is_filtered(self, document: List[str]) -> bool:
        """
        Determine if the input document should be filtered or not.

        Args:
            document (List[str]): Sentences in a training corpus to pretrain model.

        Returns:
            bool: `True` if the document should be filtered, otherwise `False`.
        """
        raise NotImplementedError()


DF = TypeVar('DF', bound=DocumentFilter)


class SequenceDocumentFilter(DocumentFilter):
    def __init__(self, document_filters: List[DF]):
        """
        Constructs a `SequenceDocumentFilter`.

        Args:
            document_filters (List[DF]): A list of DocumentFilters.
        """
        self._document_filters: List[DF] = document_filters

    def is_filtered(self, document: List[str]) -> bool:
        """
        Applies filters to the document in sequence.

        Args:
            document (List[str]): A list of sentences.

        Returns:
            bool: `True` if the document should be filtered, otherwise `False`.
        """
        return any([df.is_filtered(document) for df in self._document_filters])


class ScriptFilter(DocumentFilter):
    curly_brackets_ptn = re.compile(r'[\{|\}]')

    def is_filtered(self, document: List[str]) -> bool:
        """
        Determines if the document contains the curly bracket ```{``` to if it contains code.

        Args:
            document (List[str]): A list of sentences.

        Returns:
            bool: `True` if the document contains the curly bracket, otherwise `False`.
        """
        return any([self.curly_brackets_ptn.search(sentence) for sentence in document])


class ShortDocumentFilter(DocumentFilter):

    def __init__(self, min_sentence_num: int = 5):
        """
        Constructs a ShortDocumentFilter.

        Args:
            min_sentence_num (int): The minimum number of sentences a document should contain.
        """
        self.min_sentence_num = min_sentence_num

    def is_filtered(self, document: List[str]) -> bool:
        """
        Determines if the number of sentences in the document is suitable.

        Args:
            document (List[str]): A list of sentences.

        Returns:
            bool: `True` if the document is short, otherwise `False`.
        """
        return len(document) < self.min_sentence_num


class NGWordsFilter(DocumentFilter):
    DICT_TYPE = 'core'
    SPLIT_MODE = SplitMode.A

    def __init__(self, ng_words_file_path: str):
        """
        Constructs a NGWordsFilter.

        Args:
            ng_words_file_path (str): A file path of NG word list.
        """
        self.ng_words_file_path = ng_words_file_path
        with open(self.ng_words_file_path, 'r', encoding='utf-8') as f:
            ng_words = [line.rstrip() for line in f if line.strip() != '']
        self.ng_words_ptn = re.compile(r'({})'.format('|'.join(ng_words)))

        self.sudachi = Dictionary(dict=self.DICT_TYPE).create(self.SPLIT_MODE)

    def is_matched_by_morpheme(self, match: Match, sentence: str) -> bool:
        """
        Determines if a substring in the sentence matches at the morphological level.

        Args:
            match (Match): A Match object.
            sentence (str): A sentence.

        Returns:
            bool: `True` if a substring is included in the sentence as a word, otherwise `False`.
        """
        matched_begin_id, matched_end_id = match.span()

        morph_begin_ids = set()
        morph_end_ids = set()
        for m in self.sudachi.tokenize(sentence):
            morph_begin_id, morph_end_id = m.begin(), m.end()
            if morph_begin_id <= matched_begin_id:
                morph_begin_ids.add(morph_begin_id)
                morph_end_ids.add(morph_end_id)
            else:
                break

        return matched_begin_id in morph_begin_ids and matched_end_id in morph_end_ids

    def contain_ng_words(self, sentence: str) -> bool:
        """
        Determines if the sentence contains NG words.

        Args:
            sentence (str): A sentence.

        Returns:
            bool: `True` if the sentence contains NG words, otherwise `False`.
        """
        matches = [match for match in self.ng_words_ptn.finditer(sentence)]
        if matches:
            return any([self.is_matched_by_morpheme(match, sentence) for match in matches])
        else:
            return False

    def is_filtered(self, document: List[str]) -> bool:
        """
        Determines if the document contains even a single NG word.

        Args:
            document (List[str]): A list of sentences.

        Returns:
            bool: `True` if the document contains NG words, otherwise `False`.
        """
        return any([self.contain_ng_words(sentence) for sentence in document])



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, TypeVar


class DocumentNormalizer(object):

    def normalize(self, document: List[str]) -> List[str]:
        """
        Normalizes the document.

        Args:
            document (List[str]): A list of sentences.

        Returns:
            List[str]: A normalized document.
        """
        raise NotImplementedError()


DN = TypeVar('DN', bound=DocumentNormalizer)


class SequenceDocumentNormalizer(DocumentNormalizer):
    def __init__(self, document_normalizers: List[DN]):
        """
        Constructs a SequenceDocumentNormalizer.

        Args:
            document_normalizers (List[DN]): A list of DocumentNormalizers.
        """
        self._document_normalizers: List[DN] = document_normalizers

    def normalize(self, document: List[str]) -> List[str]:
        """
        Applies normalizers to the document in sequence.

        Args:
            document (List[str]): A list of sentences.

        Returns:
            List[str]: A normalized document.
        """
        for document_normalizer in self._document_normalizers:
            document = document_normalizer.normalize(document)

        return document


class ConcatShortSentenceNormalizer(DocumentNormalizer):

    def __init__(self, concat_char_num: int = 2):
        """
        Constructs a ConcatShortSentenceNormalizer.

        Args:
            concat_char_num (int): The maximum number of characters to be concatenated with the previous sentence.
        """
        self.concat_char_num = concat_char_num

    def normalize(self, document: List[str]) -> List[str]:
        """
        Joins a short sentence that are only a few characters to the previous sentence.

        Args:
            document (List[str]): A list of sentences.

        Returns:
            List[str]: A document with short sentences concatenated.
        """

        if len(document) == 1:
            return document
        else:
            concat_ids = []
            for i, sentence in enumerate(document):
                if 0 < i and len(sentence) <= self.concat_char_num:
                    concat_ids.append(i)

            for concat_id in concat_ids[::-1]:
                document[concat_id - 1] += document.pop(concat_id)
            return document



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
from progressbar import progressbar as tqdm
from typing import List


def write_lines(tmp_lines: List[str], file_dir: str, file_name: str, file_id: int, file_ext: str):
    with open(os.path.join(file_dir, f"{file_name}{file_id:0=2}{file_ext}"), 'w') as f:
        for line in tmp_lines:
            print(line, file=f)


def main():
    args = get_args()

    file_dir = os.path.dirname(args.input_file)
    file_name, file_ext = os.path.splitext(os.path.basename(args.input_file))

    with open(args.input_file, 'r') as f:
        lines = f.readlines()

    tmp_lines = []
    file_id = 1
    for line in tqdm(lines):
        tmp_lines.append(line.strip())
        if len(tmp_lines) > args.line_per_file and line == '\n':
            write_lines(tmp_lines, file_dir, file_name, file_id, file_ext)
            file_id += 1
            tmp_lines = []

    if len(tmp_lines) > 0:
        write_lines(tmp_lines, file_dir, file_name, file_id, file_ext)


def get_args():
    parser = argparse.ArgumentParser(description='Split dataset.')

    parser.add_argument('-i', '--input_file', help='Input file to be splitted (corpus splitted by paragraph).')
    parser.add_argument('--line_per_file', type=int, help='Max number of lines per file.')

    args = parser.parse_args()

    return args


if __name__ == '__main__':
    main()



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from enum import Enum
from typing import Callable

from sudachipy import Dictionary, Morpheme


HALF_ASCII_TRANSLATE_TABLE = str.maketrans({chr(0xFF01 + _): chr(0x21 + _) for _ in range(94)})

CONJUGATIVE_POS = {'動詞', '形容詞', '助動詞'}


class WordFormTypes(str, Enum):
    SURFACE = 'surface'
    DICTIONARY = 'dictionary'
    NORMALIZED = 'normalized'
    DICTIONARY_AND_SURFACE = 'dictionary_and_surface'
    NORMALIZED_AND_SURFACE = 'normalized_and_surface'
    SURFACE_HALF_ASCII = 'surface_half_ascii'
    DICTIONARY_HALF_ASCII = 'dictionary_half_ascii'
    DICTIONARY_AND_SURFACE_HALF_ASCII = 'dictionary_and_surface_half_ascii'
    NORMALIZED_CONJUGATION = 'normalized_conjugation'

    def __str__(self):
        return self.value


def word_formatter(word_form_type, sudachi_dict: Dictionary) -> Callable[[Morpheme], str]:
    """
    Returns the function that converts a morpheme to the specified word form.

    Args:
        word_form_type (str): Word form.
        sudachi_dict (Dictionary): Sudachi dictionary.

    Returns:
        Callable[[Morpheme], str]: The function that converts a morpheme to the specified word form.
    """

    if word_form_type not in list(WordFormTypes):
        raise ValueError('Invalid word_form_type error `{}`: {}'.format(word_form_type,
                                                                        list(map(str, WordFormTypes))))

    if word_form_type == WordFormTypes.NORMALIZED_CONJUGATION:
        from sudachitra.conjugation_preserving_normalizer import ConjugationPreservingNormalizer
        conjugation_preserving_normalizer = ConjugationPreservingNormalizer(
            os.path.join(os.path.dirname(__file__), "resources/inflection_table.json"),
            os.path.join(os.path.dirname(__file__), "resources/conjugation_type_table.json"),
            sudachi_dict)

    word_formatters = {
        WordFormTypes.SURFACE: (
            lambda m: m.surface()
        ),
        WordFormTypes.DICTIONARY: (
            lambda m: m.dictionary_form()
        ),
        WordFormTypes.NORMALIZED: (
            lambda m: m.normalized_form()
        ),
        WordFormTypes.DICTIONARY_AND_SURFACE: (
            lambda m: m.surface() if m.part_of_speech()[0] in CONJUGATIVE_POS else m.dictionary_form()
        ),
        WordFormTypes.NORMALIZED_AND_SURFACE: (
            lambda m: m.surface() if m.part_of_speech()[0] in CONJUGATIVE_POS else m.normalized_form()
        ),
        WordFormTypes.SURFACE_HALF_ASCII: (
            lambda m: m.surface().translate(HALF_ASCII_TRANSLATE_TABLE)
        ),
        WordFormTypes.DICTIONARY_HALF_ASCII: (
            lambda m: m.dictionary_form().translate(HALF_ASCII_TRANSLATE_TABLE)
        ),
        WordFormTypes.DICTIONARY_AND_SURFACE_HALF_ASCII: (
            lambda m: m.surface().translate(HALF_ASCII_TRANSLATE_TABLE) if m.part_of_speech()[0] in CONJUGATIVE_POS
            else m.dictionary_form().translate(HALF_ASCII_TRANSLATE_TABLE)
        ),
        WordFormTypes.NORMALIZED_CONJUGATION: (
            lambda m: conjugation_preserving_normalizer.normalized(m)
        )
    }

    return word_formatters[word_form_type]



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_packages, setup

setup(
    name="SudachiTra",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="Japanese tokenizer for Transformers.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/WorksApplications/SudachiTra",
    license="Apache-2.0",
    author="Works Applications",
    author_email="sudachi@worksap.co.jp",
    packages=find_packages(exclude=['tests']),
    install_requires=[
        "logzero~=1.7.0",
        "progressbar2~=3.53.1",
        "tokenizers>=0.10.3",
        "transformers>=4.6.1",
        "sudachipy>=0.6.2",
        "sudachidict_core>=20210802"
    ],
    include_package_data=True
)



# Copyright (c) 2021 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import tensorflow_datasets as tfds
from bunkai import Bunkai
from progressbar import progressbar as tqdm
from tensorflow_datasets.core import DatasetInfo
from typing import List, Tuple


TARGET_DATASETS = ['train', 'validation', 'test']

START_ARTICLE_DELIMITER = '_START_ARTICLE_'
START_PARAGRAPH_DELIMITER = '_START_PARAGRAPH_'
NEW_LINE_DELIMITER = '_NEWLINE_'


def split_article(article_text: str) -> List[List[str]]:
    """
    Splits an article into paragraphs.

    Args: article_text (str): An article in wikipedia.

    Returns:
        List[List[str]]: List of paragraphs containing sentences
    """
    paragraphs = []
    lines = article_text.split('\n')
    for i in range(2, len(lines), 2):
        if lines[i-1] == START_PARAGRAPH_DELIMITER:
            paragraphs.append(lines[i].split(NEW_LINE_DELIMITER))

    return paragraphs


def download_wiki40b_corpus(target: str) -> Tuple[DatasetInfo, List[str]]:
    """
    Downloads the target corpus and disambiguates sentence boundaries for sentences in the corpus.

    Args:
        target (str): Target dataset name.

    Returns:
        (tuple): tuple containing:
            ds_info (DatasetInfo): Dataset information for target corpus.
            all_sentences (List[str]): Sentences in the target corpus.
    """

    ds, ds_info = tfds.load(name='wiki40b/ja', split=TARGET_DATASETS, with_info=True)

    bunkai = Bunkai()

    all_sentences = []
    for line in tqdm(tfds.as_dataframe(ds[TARGET_DATASETS.index(target)], ds_info).itertuples()):
        paragraphs = split_article(line.text.decode('utf-8'))
        all_sentences.append(START_ARTICLE_DELIMITER)
        for paragraph in paragraphs:
            all_sentences.append(START_PARAGRAPH_DELIMITER)
            for sentences in paragraph:
                for sentence in bunkai(sentences):
                    if sentence:
                        all_sentences.append(sentence)

    return ds_info, all_sentences


def main():
    args = get_args()

    dataset_info, corpus_sentences = download_wiki40b_corpus(args.target)

    os.makedirs(args.output_dir, exist_ok=True)

    with open(os.path.join(args.output_dir, 'dataset_info_{}.json'.format(args.target)), 'w') as f:
        f.write(dataset_info.as_json)

    with open(os.path.join(args.output_dir, 'ja_wiki40b_{}.txt'.format(args.target)), 'w') as f:
        for sentence in corpus_sentences:
            f.write(sentence + '\n')


def get_args():
    parser = argparse.ArgumentParser(description='Download and parse target dataset.')
    parser.add_argument('-t', '--target', choices=['train', 'validation', 'test'], help='Target dataset.')
    parser.add_argument('-o', '--output_dir', required=True,
                        help='The path to the target directory in which to save a corpus file and a config file.')

    args = parser.parse_args()

    return args


if __name__ == '__main__':
    main()
